package main

import (
	"bytes"
	"crypto/x509"
	"encoding/json"
	"reflect"
	"sync"
	"time"

	//"encoding/json"
	"encoding/pem"
	"fmt"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	pb "github.com/hyperledger/fabric/protos/peer"
	"strconv"
)

type GXHCC struct {
}


type AuthGroup struct {
	Users map[string]int64 `json:"users"`  //用户组权限，默认使用用户组。我们使用形如：“User1”:timestamp. key 表示用户或者角色，value 表示有效时间。我们使用格林威治时间时间戳
	Roles map[string]int64 `json:"roles"`  //角色组权限，
}


type Auth struct {
	Read  AuthGroup `json:"read"` // 字段的读权限,“”,不存在权限，
	Write AuthGroup `json:"write"`//字段的写权限
}

type LeadgerAuth struct {
	Read  AuthGroup `json:"read"` // 字段的读权限,“”,不存在权限，
	// Write AuthGroup `json:"write"`//字段的写权限 如果是账本数据
	Issue   AuthGroup 	`json:"issue"`  //账本数据的发行权限
	Extract AuthGroup	`json:"extrate"`//账本数据的提取权限
	Transfer AuthGroup  `json:"transfer"` // 账本数据的转账权限，默认是只有账户的所属者
}

type SchemaParameters struct {
	Value interface{}  `json:"val"`//默认值  这里特别说明一下如果这里是backUp 字段，我们不进行更深一步的校验，包括格式
	PAuth  Auth         `json:"auth"`
	FCheck *Checks		`json:"fcheck"`
}


type SchemaLeadgerParameters struct {
	Value interface{}  `json:"val"`//默认值  这里特别说明一下如果这里是backUp 字段，我们不进行更深一步的校验，包括格式
	PAuth  LeadgerAuth         `json:"auth"`
	FCheck *Checks		`json:"fcheck"`
}

type Checks struct {
	FType 		string  `json:"ftype"` //字段类型
	Must  		bool 	`json:"must"`	 //是否必填
	Fieldlen  	int 	`json:"fieldlen"` //字段长度（如果字段类型是string,表示字段长度，如果是数值类型用0,1表示数据正负）
}


type Schema struct {
	//Fields 		map[string]SchemaParameters `json:"fields"`
	LeadgerFields 	map[string]SchemaLeadgerParameters `json:"leadger_fields"`  //账本字段
	StorageFields 	map[string]SchemaParameters `json:"storage_fields"`   //存证字段
	SchemaAuth 	    Auth `json:"schema_auth"`//schema 的写权限，主要是用在 schema update 部分
	SAuth 			Auth `json:"auth"`
}

type Action struct {
	AccountType 	string 		`json:"account_type"`//账户类型
	AccountId  		string		`json:"account_id"`  //账户Id
	ActionAmount 	float64		`json:"action_amount"`//这次的动作金额
	FundType        string		`json:"fund_type"`  //资金类型
	BackUp          interface{}	`json:"back_up"`//备注信息
}

type TransferAccount struct {
	AccountType 	string 		`json:"account_type"`//账户类型
	AccountId  		string		`json:"account_id"`  //账户Id
	FundType        string		`json:"fund_type"`  //资金类型
}
type Node struct {
	Key string
	Value []byte
	pre *Node
	next *Node
}


type LRUCache struct {
	limit int
	HashMap map[string]*Node
	head *Node
	end *Node
}


type Merchantsale struct {
	Mid 	  string  `json:"mid"`  //商户Id
	Bid       string  `json:"bid"`  // 分支机构id,也是一笔交易的接受者
	Tamount   float64  `json:"tamount"` //交易金额
	Treposit  float64  `json:"treposit"` //扣除预付金
	Bereposit float64  `json:"bereposit"` //扣除前预付金
	Afreposit float64  `json:"afreposit"` //扣除后预付金
	Feerate   float64  `json:"feerate"`   // 手续费扣率
	Fee		  float64  `json:"fee"`    // 手续费
	Oamount   float64  `json:"oamount"` // 原始交易金额
	Adiscounts float64 `json:"adiscounts"` //代理结算折扣
	Sdiscounts float64 `json:"sdiscounts"` //消费折扣
	Nodisamount float64 `json:"nodisamount"` //不打折金额
	Disamount   float64 `json:"disamount"`  //打折金额
	Perquisites float64 `json:"perquisites"` //额外收益
}


type Merchant struct {
	Id     		string 		`json:"id"`

}

type CommonTx struct {
	SenderAccountType  	string 	`json:"sender_account"`
	SenderId			string  `json:"sender_id"`
	SenderFunt		    string	`json:"sender_funt"`
	ReceiverAccountType string	`json:"receiver_account"`
	ReceiverId  		string  `json:"receiver_account_id"`
	ReceiverFunt  		string	`json:"receiver_funt"`
	Amount				float64 `json:"amount"`
}

type transferTypes func(args []string, stub shim.ChaincodeStubInterface)error // 声明了一个函数类型

var (
	defValue = "***" //查询用户权限不足的情况下默认返回值
	PREFIX = "GXH"
	StoragePrefix = "SGXH"
	LeadgerPrefix = "LGXH"
	IsUsers = true   //是否启用用户权限，false 表示启用Roles
	GxhSchema = "gxhSchema" //schema 存储前缀，为了实现复合主键使用。
	InitSchemaLength = 12 //初始化时schema 的长度
	schemaCache = Constructor(InitSchemaLength)
	once = sync.Once{}
	transferType = initTransfer()
)


// Init does nothing for this cc
func (t *GXHCC) Init(stub shim.ChaincodeStubInterface) pb.Response {
	return shim.Success([]byte("GXHCC init successful! "))
}

/*
	初始化我们要使用的交易类型
*/
func initTransfer()map[string]transferTypes{
	transferType := make(map[string]transferTypes)
	transferType["merchants"] = merchantsale
	transferType["tickerorders"] = tickerorders
	transferType["commontx"] = normalTransfer
	return transferType
}

func initSchema(stub shim.ChaincodeStubInterface)error{
	var errinfo error
	once.Do(func(){
		schemaIterator,err :=  stub.GetStateByPartialCompositeKey(GxhSchema,nil)
		if err != nil {
			errinfo = err
			return
		}
		defer schemaIterator.Close()
		index := 1
		for schemaIterator.HasNext(){
			querykv,err := schemaIterator.Next()
			if err != nil {
				errinfo = err
				return
			}
			fmt.Println("add cache key:",querykv.Key)
			_,compisiteKeys ,err := stub.SplitCompositeKey(querykv.Key)
			if err!=nil{
				errinfo = err
				return
			}
			schemaCache.Put(compisiteKeys[0],querykv.Value)
			fmt.Println("add cache key:",querykv.Key)
			index++
			if index > InitSchemaLength {
				fmt.Println("cache data length :",len(schemaCache.HashMap))
				return
			}
		}
		fmt.Println("cache data length :",len(schemaCache.HashMap))
		return
	})
	if errinfo != nil{
		return errinfo
	}
	return nil
}

// Invoke for this chaincode exposes
func (t *GXHCC) Invoke(stub shim.ChaincodeStubInterface) pb.Response {
	// get arguments and transient
	functionName, args := stub.GetFunctionAndParameters()
	err := initSchema(stub)
	if err !=nil{
		return shim.Error(err.Error())
	}
	switch functionName {
	case "authority":

	case "schema":
		res,err := SchemaProcess(args,stub)
		if err!=nil{
			return shim.Error(fmt.Sprintf("SchemaProcess function err: %s", err))
		}
		return shim.Success([]byte(res))
	case "putData":
		res,err := putData(args,stub)
		if err!=nil{
			return shim.Error(fmt.Sprintf("putData function err: %s", err))
		}
		return shim.Success([]byte(res))
	case "get":
		res,err := get(args,stub)
		if err!=nil{
			return shim.Error(fmt.Sprintf("get function err: %s", err))
		}
		return shim.Success([]byte(res))
	case "upStoreData":  //
		res,err := updateStoreData(args,stub)
		if err!=nil{
			return shim.Error(fmt.Sprintf("upStoreData function err: %s", err))
		}
		return shim.Success([]byte(res))
	case "createAccount":
		res,err := createAccount(args,stub)
		if err!=nil{
			return shim.Error(fmt.Sprintf("createAccount function err: %s", err))
		}
		return shim.Success([]byte(res))
	case "createTx":
		res,err := transactionProcess(args,stub)
		if err!=nil{
			return shim.Error(fmt.Sprintf("transactionProcess function err: %s", err))
		}
		return shim.Success([]byte(res))
	case "getCertCommonName":
		name,err := getCertificateCommonName(stub)
		if err != nil{
			return shim.Error(fmt.Sprintf("get certificate CommonName fail,err: %s",err))
		}
		return shim.Success([]byte("current user certificate commonname :" +name))
	default:
		return shim.Error(fmt.Sprintf("Unsupported function %s", functionName))
	}
	return shim.Success([]byte("GXHCC invoke successful "))
}

/*
	根据账户类型创建账本账户，由于我们这里有存储数据与账本数据两种，所以我们还需要在schema 部分标识不同的数据类型。
	伪代码：
	1：检查当前用户是否具有账户的总写（创建）权限
	2：检查账户必备的属性，如Id,账户名，账户所属者，对应的是证书commonName
	2：根据schema 标注的金额（账本）字段，和存证字段，分别进行数据的putState存储工作。其中在存证数据的Id 字段是标示字段。
	3：完成账户创建工作
*/
func createAccount(args []string, stub shim.ChaincodeStubInterface) (string, error){
	if len(args) < 2 {
		return "", fmt.Errorf("put data operation expected more than 2 parameters! ")
	}
	var PutDatas []map[string]interface{}
	commonName, err := getCertificateCommonName(stub)
	if err != nil {
		return "", fmt.Errorf("getCertificateCommonName happen err: %s ", err)
	}
	schema := &Schema{}
	err = schema.get(args[0], stub)
	if err != nil {
		return "", err
	}
	if ok,err := authorityCheck(schema.SAuth.Write, commonName);!ok {
		return "", fmt.Errorf("write %s table fail,err: %s ", args[0],err)
	}
	if err := json.Unmarshal([]byte(args[1]), &PutDatas); err != nil {
		return "", fmt.Errorf("%s table parameters Unmarshal fail,put string not json array, err:%s",args[0], err)
	}
	for k, mapres := range PutDatas {
		Id, ok := mapres["id"]
		if !ok {
			return "", fmt.Errorf("The id field must exist ")
		}
		_, ok = mapres["common_name"]
		if !ok {
			return "", fmt.Errorf("The common_name field must exist ")
		}
		err = SchemaCheck(schema, mapres)
		if err != nil {
			return "", fmt.Errorf("args index %d string, %s schema cheack fail,err : %s ", k, args[0],err)
		}
		if err := FiledsCheck(schema,mapres);err !=nil{
			return "",err
		}
		storagemap,leadgermap := separateFields(schema,mapres,[]string{"id","common_name"})
		storageDate, _ := json.Marshal(storagemap)
		leadgerDate, _ := json.Marshal(leadgermap)
		key := StoragePrefix+"_"+args[0]+"_"+Id.(string)
		getResult,err := stub.GetState(key)
		if err != nil {
			return "", fmt.Errorf("GetState %s data fail,please check your parameters ",key)
		}
		if getResult != nil{
			return "", fmt.Errorf("%s data already exists and cannot be added",key)
		}
		if err := stub.PutState(key, storageDate); err != nil {
			return "", fmt.Errorf("PutState fail, err :%s", err)
		}
		if err := stub.PutState(LeadgerPrefix+"_"+args[0]+"_"+Id.(string), leadgerDate); err != nil {
			return "", fmt.Errorf("PutState fail, err :%s", err)
		}
	}
	return fmt.Sprintf("create %s account success!", args[0]), nil
}


/*
	修改一个账户的存储数据，修改规则由Schema 控制,有些特殊数据在账本数据也使用，所以无法修改
*/
func updateStoreData(args []string, stub shim.ChaincodeStubInterface) (string, error) {
	if len(args) != 2 {
		return "", fmt.Errorf("update data operation expected 2 parameters! ")
	}
	commonName, err := getCertificateCommonName(stub)
	if err != nil {
		return "", fmt.Errorf("getCertificateCommonName happen err: %s ", err)
	}
	schema := &Schema{}
	err = schema.get(args[0], stub)
	if err != nil {
		return "", err
	}
	var Updata map[string]interface{}
	if err := json.Unmarshal([]byte(args[1]), &Updata); err != nil {
		return "", fmt.Errorf("%s table parameters Unmarshal fail,err: %s",args[0], err)
	}
	Id, ok := Updata["id"]
	if !ok {
		return "", fmt.Errorf("The id field must exist ")
	}
	_, ok = Updata["common_name"]
	if ok {
		return "", fmt.Errorf("The common_name field cannot update ")
	}
	if ok ,_ := authorityCheck(schema.SAuth.Write, commonName);!ok {
		if ok, err := parsauthorityCheck(schema.StorageFields, Updata, commonName); !ok {
			return "", fmt.Errorf("authority Check err: %s", err)
		}
	}
	if err := FiledsCheck(schema,Updata);err !=nil{
		return "",err
	}
	key := StoragePrefix+"_"+args[0]+"_"+Id.(string)
	allMap,err := getState(key,stub)
	if err != nil  {
		return "", err
	}

	if err := Deassign(Updata, allMap); err != nil {
		return "", err
	}
	err = putState(key,allMap,stub)
	if err != nil  {
		return "", err
	}
	return fmt.Sprintf("%s update data success!", args[0]), nil
}

/*
	存储数据，某些不产生账本数据的数据使用这个接口来存储。例如码数据，进件表数据
	允许覆盖
*/
func putData(args []string, stub shim.ChaincodeStubInterface) (string, error) {
	if len(args) < 2 {
		return "", fmt.Errorf("put data operation expected more than 2 parameters! ")
	}
	var PutDatas []map[string]interface{}
	commonName, err := getCertificateCommonName(stub)
	if err != nil {
		return "", fmt.Errorf("getCertificateCommonName happen err: %s ", err)
	}
	schema := &Schema{}
	err = schema.get(args[0], stub)
	if err != nil {
		return "", err
	}
	if ok,err := authorityCheck(schema.SAuth.Write, commonName);!ok {
		return "", fmt.Errorf("write %s table fail,err: %s ", args[0],err)
	}
	if err := json.Unmarshal([]byte(args[1]), &PutDatas); err != nil {
		return "", fmt.Errorf("%s table parameters Unmarshal fail,put string not json array, err:%s",args[0], err)
	}
	for k, mapres := range PutDatas {
		Id, ok := mapres["id"]
		if !ok {
			return "", fmt.Errorf("The id field must exist ")
		}
		if len(schema.LeadgerFields)!= 0{
			return "", fmt.Errorf("%s schema exist leadger fields,cannot putData", args[0])
		}
		err = SchemaCheck(schema, mapres)
		if err != nil {
			return "", fmt.Errorf("args index %d string, %s schema cheack fail,err : %s ", k, args[0],err)
		}
		if err := FiledsCheck(schema,mapres);err !=nil{
			return "",err
		}
		key := StoragePrefix+"_"+args[0]+"_"+Id.(string)
		err = putState(key,mapres,stub)
		if err != nil  {
			return "", err
		}
	}
	return fmt.Sprintf("%s put data success!", args[0]), nil
}

/*
	查询数据，如果存在账本数据，与存证数据，则进行数据拼接，然后返回。
*/
func get(args []string, stub shim.ChaincodeStubInterface) (string, error) {
	if len(args) != 2 {
		return "", fmt.Errorf("get data operation expected  2 parameters to function get!")
	}
	commonName, err := getCertificateCommonName(stub)
	if err != nil {
		return "", fmt.Errorf("getCertificateCommonName happen err: %s ", err)
	}
	var parameters map[string]interface{}
	json.Unmarshal([]byte(args[1]), &parameters)
	if err != nil {
		return "", fmt.Errorf("parameters Unmarshal fail,err: %s", err)
	}

	Id, ok := parameters["id"]
	if !ok {
		return "", fmt.Errorf("The id field must exist ")
	}
	storageKey := StoragePrefix+"_"+args[0]+"_"+Id.(string)
	mapResult, err := getState(storageKey,stub)
	if err != nil {
		return "", err
	}
	leadgerKey := LeadgerPrefix+"_"+args[0]+"_"+Id.(string)
	leadgerResult, err := stub.GetState(leadgerKey)
	if err != nil {
		return "", fmt.Errorf("get %s data fail,please check your parameters ",leadgerKey)
	}
	schema := &Schema{}
	err = schema.get(args[0], stub)
	if err != nil {
		return "", err
	}
	if leadgerResult != nil {
		if err := json.Unmarshal(leadgerResult, &mapResult); err != nil {
			return "", fmt.Errorf("The leadger original data Unmarshal fail err:%s ", err)
		}
	}
	if ok,_:= authorityCheck(schema.SAuth.Read, commonName);!ok {
		if filteredData, err := dataFilter(schema, mapResult, commonName); err != nil {
			return "", fmt.Errorf("get data authority filter fail, err: %s", err)
		} else {
			result, _ := json.Marshal(filteredData)
			return string(result), nil
		}

	} else {
		result, _ := json.Marshal(mapResult)
		return string(result), nil
	}
}

func SchemaCheck(schema *Schema, checkmap map[string]interface{}) error {
	if len(schema.StorageFields)+len(schema.LeadgerFields) != len(checkmap) {
		return fmt.Errorf("The input field length is less than the schema field length ! ")
	}
	for k, v := range schema.StorageFields {
		if schemaV, ok := checkmap[k]; !ok {
			return fmt.Errorf("There is no input storage field %s ",k)
		} else {
			if parameterS, ok := v.Value.(map[string]interface{}); ok {
				if parameterC, ok := schemaV.(map[string]interface{}); !ok {
					return fmt.Errorf("The input field %s does not match the output field format ",k)
				}else{
					err := BackUpCheck(parameterS,parameterC)
					if err != nil{
						return err
					}
				}
			}
		}
	}
	for k, _ := range schema.LeadgerFields {
		if _, ok := checkmap[k]; !ok {
			return fmt.Errorf("There is no input leadger field %s ",k)
		}
	}
	return nil
}

func FiledsCheck(schema *Schema, checkmap map[string]interface{}) error{
	for k, schemaV := range schema.StorageFields {
		if k != "backup" && schemaV.FCheck!= nil && schemaV.FCheck.FType != "" && checkmap[k] != nil{
			if err := fieldCheckout(schemaV.FCheck,checkmap[k]);err!=nil{
				return fmt.Errorf("%s %s ",k,err)
			}
		}
	}
	for k, schemaV := range schema.LeadgerFields {
		if schemaV.FCheck!= nil && schemaV.FCheck.FType != "" && checkmap[k] != nil{
			if err := fieldCheckout(schemaV.FCheck,checkmap[k]);err!=nil{
				return fmt.Errorf("%s %s ",k,err)
			}
		}
	}
	return nil
}

func BackUpCheck(backUp map[string]interface{}, checkmap map[string]interface{}) error {
	if len(backUp) != len(checkmap) {
		return fmt.Errorf("The input backup field length is less than the schema backup field length ! ")
	}
	for k, _ := range checkmap {
		if _, ok := backUp[k]; !ok {
			return fmt.Errorf("There is no input backup field %s ",k)
		}
	}
	return nil
}

/*
	将字段根据账本数据，与存储数据分开
	leadgers:参数表示我们需要在账本数据包含的字段，这些字段不属于账本数据。
*/
func separateFields(schema *Schema,checkmap map[string]interface{},leadgers []string)(storagemap ,leadgermap map[string]interface{}){
	storagemap = make(map[string]interface{})
	leadgermap = make(map[string]interface{})
	for k,_ :=range schema.LeadgerFields{
		leadgermap[k] = checkmap[k]
	}
	for _,v :=range leadgers{
		leadgermap[v] = checkmap[v]
	}
	for k,_ :=range schema.StorageFields{
		storagemap[k] = checkmap[k]
	}
	return
}

/*
	根据权限拼装数据,进行数据过滤
*/
func dataFilter(schema *Schema, checkmap map[string]interface{}, commonName string) (map[string]interface{}, error) {
	for k, schemaV := range schema.StorageFields {
		if v, ok := checkmap[k]; !ok {
			return nil, fmt.Errorf("Update data parameter mismatch with schema structure! ")
		} else {
			if _, ok := schemaV.Value.(map[string]SchemaParameters); ok {
				if _, ok := v.(map[string]interface{}); !ok {
					return nil, fmt.Errorf("Update data parameter mismatch with schema structure! ")
				}
				if ok,_ := authorityCheck(schemaV.PAuth.Read, commonName);!ok {
					checkmap[k] = defValue
				}
			} else {
				if ok,_ := authorityCheck(schemaV.PAuth.Read, commonName);!ok {
					checkmap[k] = defValue
				}
			}
		}
	}
	for k, schemaV := range schema.LeadgerFields {
		if v, ok := checkmap[k]; !ok {
			return nil, fmt.Errorf("Update data parameter mismatch with schema structure! ")
		} else {
			if _, ok := schemaV.Value.(map[string]SchemaLeadgerParameters); ok {
				if _, ok := v.(map[string]interface{}); !ok {
					return nil, fmt.Errorf("Update data parameter mismatch with schema structure! ")
				}
				if ok,_ := authorityCheck(schemaV.PAuth.Read, commonName);!ok {
					checkmap[k] = defValue
				}
			} else {
				if ok,_ := authorityCheck(schemaV.PAuth.Read, commonName);!ok {
					checkmap[k] = defValue
				}
			}
		}
	}
	return checkmap, nil
}


func SchemaProcess(args []string, stub shim.ChaincodeStubInterface) (string, error) {
	if len(args) < 2 {
		return "", fmt.Errorf("Expect features with 2 or more parameters to play! ")
	}
	schema := &Schema{}
	switch args[0] {

	case "put":
		return schema.put(args[1:], stub)
	case "update":
		return schema.update(args[1:], stub)
	case "get":
		return schema.getSchema(args[1], stub)

	default:
		return "", fmt.Errorf(fmt.Sprintf("Unsupported schema function of %s", args[0]))

	}
	return "", nil
}
func transactionProcess(args []string, stub shim.ChaincodeStubInterface) (string, error) {
	if len(args) < 2 {
		return "", fmt.Errorf("Expect features with 2 or more parameters to play! ")
	}
	switch args[0] {

	case "issue":  //进行账户金额发行（无中生有）
		return funtOperation(args[1:],stub,true)
	case "extract": //提取账户金额（直接减去）
		return funtOperation(args[1:],stub,false)
	case "transfer": // 一笔普通转账
		return transferTx(args[1:],stub)
	default:
		return "", fmt.Errorf(fmt.Sprintf("Unsupported schema function of %s", args[0]))

	}
	return "", nil
}
/*
	put
	args 字段说明：
	1：args[0]代表我们具体操作的表，
    2：args[1:]后面的数据表示要记录到数据库的数据,每一个string 必须是json字符串，并且需要每一个json 字符串都必须包含id 字段

我们可以发现json字符串都被我们解析成了map[string]interface{}类型的map,其中key代表表模板的字段，interface可以是一个结构，我们当前将他看做我们表模板字段的初始值。
伪代码：
1：在创建schema data 时，我们必须检查金额字段的类型必须是float64,并且不能为负数。
*/
func (this *Schema) put(args []string, stub shim.ChaincodeStubInterface) (string, error) {
	if len(args) != 2 {
		return "", fmt.Errorf("Schema  put operation expected  2 parameters!")
	}
	var schema *Schema
	var err error
	if schema ,err = NewSchema(args[1]);err != nil{
		return "", fmt.Errorf("The parameters Unmarshal to a Schema structural fail,err: %s ", err)
	}else{
		if backupValue, ok := schema.StorageFields["backup"]; ok {
			if _,ok := backupValue.Value.(map[string]interface{}); !ok {
				return "", fmt.Errorf("backup parameters Unmarshal fail,it not json string,err: %s", err)
			}
		}
	}
	_, ok := schema.StorageFields["id"]
	if !ok {
		return "", fmt.Errorf("The id field must exist ")
	}
	_, ok = schema.StorageFields["common_name"]
	if !ok {
		return "", fmt.Errorf("The common_name field must exist ")
	}

	CompositeKey,err := stub.CreateCompositeKey(GxhSchema,[]string{args[0]})
	if err != nil{
		return "", fmt.Errorf("the schema %s create comosite key fail,err : %s", args[0],err)
	}
	if schemaCache.Get(args[0]) != nil {
		fmt.Println("schemaCache.Get err")
		return "", fmt.Errorf("the schema %s already exists and cannot be added", args[0])
	}else if value, _ := stub.GetState(CompositeKey); value != nil {
		fmt.Println("GetState.Get err")
		return "", fmt.Errorf("the schema %s already exists and cannot be added", args[0])
	}
	putDate, _ := json.Marshal(schema) //处理重复字段
	if err := stub.PutState(CompositeKey, putDate); err != nil {
		return "", fmt.Errorf("Schema PutState fail,err: %s ", err)
	}
	schemaCache.Put(args[0],putDate)
	return fmt.Sprintf("%s schema put success!", args[0]), nil
}

/*
	if update Schema 那么我们没有办法比较旧有的一个模板，
     Schema 的 update,不支持部分修改，因为存在增加和删除某些字段的情况。
     初版暂时只允许修改存储数据,金额字段暂时不允许修改
*/
func (this *Schema) update(args []string, stub shim.ChaincodeStubInterface) (string, error) {
	if len(args) != 2 {
		return "", fmt.Errorf("Schema update operation expected 2 parameters! ")
	}
	var schema *Schema
	var schemaOld *Schema
	var err error
	if schema ,err = NewSchema(args[1]);err != nil{
		return "", fmt.Errorf("The parameters Unmarshal to a Schema structural fail,err: %s ", err)
	}else{
		if backupValue, ok := schema.StorageFields["backup"]; ok {
			if _,ok := backupValue.Value.(map[string]interface{}); !ok {
				return "", fmt.Errorf("backup parameters Unmarshal fail,it not json string,err: %s", err)
			}
		}
	}
	_, ok := schema.StorageFields["id"]
	if !ok {
		return "", fmt.Errorf("The id field must exist ")
	}
	_, ok = schema.StorageFields["common_name"]
	if !ok {
		return "", fmt.Errorf("The common_name field must exist ")
	}
	CompositeKey,err := stub.CreateCompositeKey(GxhSchema,[]string{args[0]})
	if err != nil{
		return "", fmt.Errorf("the schema %s create comosite key fail,err : %s", args[0],err)
	}
	var result []byte
	if result = schemaCache.Get(args[0]);result == nil{
		fmt.Println("schemaCache.Get nil")
		result, err := stub.GetState(CompositeKey)
		if err != nil || result == nil {
			return "", fmt.Errorf("%s schema data doesn't exist,please check your parameters ",args[0])
		}
	}
	if err := json.Unmarshal(result, &schemaOld); err != nil {
		return "", fmt.Errorf("The parameters Unmarshal to a Schema structural fail, err: %s ", err)
	}
	commonName, err := getCertificateCommonName(stub)
	if err != nil {
		return "", fmt.Errorf("getCertificateCommonName happen err: %s ", err)
	}
	if ok,err := authorityCheck(schemaOld.SchemaAuth.Write,commonName);!ok{
		return "", fmt.Errorf("update %s schema fail,err: %s ",args[0],err)
	}
	schema.LeadgerFields = schemaOld.LeadgerFields
	updateData,_ := json.Marshal(schema)
	if err = stub.PutState(CompositeKey, updateData);err != nil{
		return "", fmt.Errorf("%s schema data PutState fail,please try again ",args[0])
	}
	schemaCache.Put(args[0],updateData)
	return fmt.Sprintf("%s schema update success!", args[0]), err

}

func (this *Schema) get(args string, stub shim.ChaincodeStubInterface) error {
	var result []byte
	if result = schemaCache.Get(args);result == nil{
		CompositeKey,err := stub.CreateCompositeKey(GxhSchema,[]string{args})
		if err != nil{
			return  fmt.Errorf("the schema %s create comosite key fail,err : %s", args,err)
		}
		result, err = stub.GetState(CompositeKey)
		if err != nil || result == nil {
			return fmt.Errorf("%s schema data doesn't exist,please check your parameters ",args)
		}
		schemaCache.Put(args,result)
	}
	return json.Unmarshal(result,this)
}

func (this *Schema) getSchema(args string, stub shim.ChaincodeStubInterface)(string, error) {
	var result []byte
	if result = schemaCache.Get(args);result == nil{
		CompositeKey,err := stub.CreateCompositeKey(GxhSchema,[]string{args})
		if err != nil{
			return "", fmt.Errorf("the schema %s create comosite key fail,err : %s", args,err)
		}
		result, err = stub.GetState(CompositeKey)
		if err != nil || result == nil {
			return "", fmt.Errorf("%s schema data doesn't exist,please check your parameters ",args)
		}
		schemaCache.Put(args,result)
	}
	if err := json.Unmarshal(result, this); err != nil {
		return "", fmt.Errorf("SchemaData Unmarshal fail, err: %s", err)
	}
	commonName, err := getCertificateCommonName(stub)
	if err != nil {
		return "", fmt.Errorf("getCertificateCommonName happen err: %s ", err)
	}
	if ok,err := authorityCheck(this.SchemaAuth.Read,commonName);!ok{
		return "",fmt.Errorf("get %s schema fail,err: %s", args,err)
	}
	return string(result), nil
}


func transferTx(args []string, stub shim.ChaincodeStubInterface)(string, error) {
	txFunc := transferType[args[0]]
	if txFunc == nil{
		return "", fmt.Errorf("No matching transfer operation method was found ! ")
	}
	err := txFunc(args[1:],stub)
	if err != nil {
		return "", fmt.Errorf("%s transfer operation fail,err:%s ",args[0],err)
	}
	return "", nil
}

func authorityCheck(authority AuthGroup, commonName string) (bool,error) {
	var auths map[string]int64
	if IsUsers {
		auths = authority.Users
	}else{
		auths = authority.Roles
	}
	if len(auths) == 0 {
		return true,nil
	}
	for k, v := range auths {
		if k == commonName {
			if v == 0 || v > time.Now().Unix(){
				return true,nil
			}else{
				return false,fmt.Errorf("%s user permission period ",commonName)
			}
		}
	}
	return false,fmt.Errorf("%s user does not have permission",commonName)

}

/*
	检查一组参数的权限是否匹配
*/
func parsauthorityCheck(schema map[string]SchemaParameters, checkmap map[string]interface{}, commonName string) (bool, error) {
	for k, v := range checkmap {
		if schemaV, ok := schema[k]; !ok {
			return false, fmt.Errorf("Update data parameter mismatch with schema structure! ")
		} else {
			if _, ok := schemaV.Value.(map[string]interface{}); ok {
				if _, ok := v.(map[string]interface{}); !ok {
					return false, fmt.Errorf("Update data parameter mismatch with schema structure! ")
				}
			}
			if ok,err := authorityCheck(schemaV.PAuth.Write, commonName);!ok {
				return false, fmt.Errorf("%s field permission check does not match, err: %s ", k,err)
			}
		}
	}
	return true, nil
}

/*
	重新赋值，并检查相应的格式是否正确
*/
func Deassign(upMap, result map[string]interface{}) error {
	if len(upMap) > len(result) {
		return fmt.Errorf("Deassign fail,reassigned data length greater than the original data length! ")
	}
	for k, v := range upMap {
		if resultV, ok := result[k]; ok {
			if resultMap, resultok := resultV.(map[string]interface{}); resultok {
				if vMap, ok := v.(map[string]interface{}); !ok {
					return fmt.Errorf("Deassign fail, format doesn't match! ")
				} else {
					Deassign(vMap, resultMap)
				}
			} else  {
				result[k] = v
			}
		} else {
			return fmt.Errorf("Deassign fail %s key nonentity ", k)
		}
	}
	return nil
}



/*
	账户金额发现，核心是在对应的schema 找到有发行的权利。
	1：金额字段也就是（账本字段数据）都有发行标示
	2：发行交易可以存在多个字段，但只要保证我们要求的字段存在就可以。账户类型，Id，发行金额（大于0）
*/
func funtOperation(args []string, stub shim.ChaincodeStubInterface,isIssue bool) (string, error){
	if len(args) != 1 {
		return "", fmt.Errorf("issue or extract  operation expected 1 parameters ! ")
	}
	commonName, err := getCertificateCommonName(stub)
	if err != nil {
		return "", fmt.Errorf("getCertificateCommonName happen err: %s ", err)
	}
	action := &Action{}
	err = json.Unmarshal([]byte(args[0]),action)
	if err != nil{
		return "", fmt.Errorf("json unmarshal Action structure fail,err: %s",err)
	}
	schema := &Schema{}
	err = schema.get(action.AccountType, stub)
	if err != nil {
		return "", err
	}
	if isIssue{
		if ok,err := authorityCheck(schema.LeadgerFields[action.FundType].PAuth.Issue, commonName);!ok {
			return "", fmt.Errorf("%s field issue permission check does not match, err: %s ", action.FundType , err)
		}
	}else {
		if ok,err := authorityCheck(schema.LeadgerFields[action.FundType].PAuth.Extract, commonName);!ok {
			return "", fmt.Errorf("%s field extract permission check does not match, err: %s ", action.FundType , err)
		}
	}

	leadgerKey := LeadgerPrefix+"_"+action.AccountType+"_"+action.AccountId
	leadgerResult, err := stub.GetState(leadgerKey)
	if err != nil {
		return "", fmt.Errorf("GetState %s data fail,err %s ",leadgerKey,err)
	}
	if leadgerResult == nil {
		return "", fmt.Errorf("GetState %s data is null,please check your parameters ",leadgerKey)
	}
	var leadgermap map[string]interface{}
	err = json.Unmarshal(leadgerResult,&leadgermap)
	if err != nil{
		return "", fmt.Errorf("json unmarshal leadger map fail,err: %s",err)
	}
	funt := leadgermap[action.FundType].(float64)
	amount := Decimal(action.ActionAmount)
	if isIssue {
		funt += amount
	}else{
		if funt < amount{
			return "", fmt.Errorf("The current amount is less than the withdrawal amount and cannot be extracted ")
		}
		funt-= amount
	}
	leadgermap[action.FundType] = funt
	leadgerResult,_ = json.Marshal(leadgermap)
	err = stub.PutState(leadgerKey,leadgerResult)
	if err!=nil{
		return "", fmt.Errorf("putState %s data fail,err: %s",leadgerKey,err)
	}
	return "transaction successful!", nil
}

/*

*/
func normalTransfer(args []string, stub shim.ChaincodeStubInterface)error{
	if len(args) != 1 {
		return nil
	}
	tx := &CommonTx{}
	err := json.Unmarshal([]byte(args[0]),tx)
	if err!= nil{
		fmt.Errorf("CommonTx structure json unmarshal fail,err : %s ",err)
	}
	sender := &TransferAccount{
		tx.SenderAccountType,
		tx.SenderId,
		tx.SenderFunt,
	}
	receiver := &TransferAccount{
		tx.ReceiverAccountType,
		tx.ReceiverId,
		tx.ReceiverFunt,
	}
	commonName ,err := getCertificateCommonName(stub)
	if err!= nil {
		return nil
	}
	return transfer(Decimal(tx.Amount),sender,receiver,commonName,stub)
}




/*
	当增加一条交易记录的时候，我们需要根据交易记录实现，1.预付金的计算，2.计算分成，3.预付金的变化。分成的计算
*/
func merchantsale(args []string, stub shim.ChaincodeStubInterface)error{
	commanName,err := getCertificateCommonName(stub)
	if err!=nil{
		return err
	}
	senderaccountType := "merchants"
	receiverAccountType := "alibusi"
	mer := &Merchantsale{}
	err = json.Unmarshal([]byte(args[0]),mer)
	if err != nil {
		return fmt.Errorf("Merchantsale struct json unmarshal fail,err : %s ",err)
	}
	//1.计算预付金 这部分待定1698.83
	var treposit float64
	if mer.Nodisamount == 0 || mer.Disamount == 0{ //如果打折金额与不打折金额任一为0，则认为全部打折
		treposit = Decimal(mer.Oamount * mer.Adiscounts)
	}else{
		treposit = Decimal(mer.Nodisamount+ mer.Disamount * mer.Adiscounts)
	}
	if treposit != mer.Treposit {
		return fmt.Errorf("treposit calculation fail ")
	}
	if mer.Fee != Decimal(mer.Tamount * mer.Feerate){
		return fmt.Errorf("fee calculation fail ")
	}
	profit := mer.Tamount - mer.Fee - treposit //计算利润
	if profit < 0 {
		return fmt.Errorf("profit is negative ")
	}
	sender := &TransferAccount{
		senderaccountType,
		mer.Mid,
		"regular",
	}
	receiver := &TransferAccount{
		receiverAccountType,
		mer.Bid,
		"abail",
	}
	err = transfer(treposit,sender,receiver,commanName,stub)
	if err!=nil{
		return nil
	}

	//进行分成数据

	return nil
}

/*
	共享惠订单交易
*/
func tickerorders(args []string, stub shim.ChaincodeStubInterface)error{
	/*
		由于共享惠还没有进行确定的订单交易所以这部分先进行伪代码的记录
	1：计算实际支付是否正确，
	2：计算手续费总和是否正确
	3：计算利润
	4：是否需要进行商户金额的清分（待定）
	5：合作方分成
	6：最终分成
	*/
	return nil
}

/*
	实现普通的转账操作
*/
func transfer(amount float64,sender ,receiver *TransferAccount,commonName string,stub shim.ChaincodeStubInterface)error{
	if amount <= 0 {
		return fmt.Errorf("transfer limit must greater than 0 ")
	}
	senderSchema := &Schema{}
	err := senderSchema.get(sender.AccountType, stub)
	if err != nil {
		return err
	}
	if ok,err := authorityCheck(senderSchema.LeadgerFields[sender.FundType].PAuth.Transfer, commonName);!ok {
		return fmt.Errorf("%s field issue permission check does not match, err: %s ", sender.FundType , err)
	}
	senderKey := LeadgerPrefix+"_"+sender.AccountType+"_"+sender.AccountId
	senderleadgermap,err := getState(senderKey,stub)
	if err != nil{
		return err
	}
	funt := senderleadgermap[sender.FundType].(float64)
		if funt < amount{
			return  fmt.Errorf("The current amount is less than the withdrawal amount and cannot be extracted ")
		}
		funt-= amount
	senderleadgermap[sender.FundType] = funt

	err = putState(senderKey,senderleadgermap,stub)
	if err != nil{
		return err
	}
	receiverKey := LeadgerPrefix+"_"+receiver.AccountType+"_"+receiver.AccountId
	receiverleadgermap,err := getState(receiverKey,stub)
	if err != nil{
		return err
	}
	rfunt := receiverleadgermap[receiver.FundType].(float64)
	rfunt+= amount
	receiverleadgermap[receiver.FundType] = rfunt

	err = putState(receiverKey,receiverleadgermap,stub)
	if err != nil{
		return err
	}
	return nil
}


func getState(key string,stub shim.ChaincodeStubInterface)(map[string]interface{},error){
	var resultmap map[string]interface{}
	result, err := stub.GetState(key)
	if err != nil {
		return nil, fmt.Errorf("GetState %s data fail,err %s ",key,err)
	}
	if result == nil {
		return  nil,fmt.Errorf("GetState %s data is null,please check your parameters ",key)
	}
	err = json.Unmarshal(result,&resultmap)
	if err != nil{
		return nil,fmt.Errorf("json unmarshal leadger map fail,err: %s",err)
	}
	return  resultmap,nil
}

func putState(key string,value map[string]interface{},stub shim.ChaincodeStubInterface)error{
	valueByte,err := json.Marshal(value)
	if err != nil{
		return fmt.Errorf(" %s json marshal data fail,err: %s",key,err)
	}
	err = stub.PutState(key,valueByte)
	if err!=nil{
		return fmt.Errorf("putState %s data fail,err: %s",key,err)
	}
	return nil
}

func unmarshalMap(arg string)(error,map[string]interface{}){
	var valueMap map[string]interface{}
	err := json.Unmarshal([]byte(arg),&valueMap)
	if err != nil{
		return err,nil
	}
	return nil,valueMap
}

func getCertificateCommonName(stub shim.ChaincodeStubInterface) (string, error) {
	return "Admin@org2.example.com", nil
	creatorByte, _ := stub.GetCreator()
	certStart := bytes.IndexAny(creatorByte, "-----BEGIN")
	if certStart == -1 {
		return "", fmt.Errorf("No certificate found ")
	}
	certText := creatorByte[certStart:]
	bl, _ := pem.Decode(certText)
	if bl == nil {
		return "", fmt.Errorf("Could not decode the PEM structure ")
	}
	cert, err := x509.ParseCertificate(bl.Bytes)
	if err != nil {
		return "", fmt.Errorf("ParseCertificate failed")
	}
	return cert.Subject.CommonName, nil
}

func Decimal(value float64) float64 {
	value, _ = strconv.ParseFloat(fmt.Sprintf("%.2f", value), 64)
	return value
}

func NewSchema(arg string)(*Schema,error){
	schema := &Schema{
		LeadgerFields:make(map[string]SchemaLeadgerParameters),
		StorageFields:make(map[string]SchemaParameters),
	}
	var argMap map[string]interface{}
	if err := json.Unmarshal([]byte(arg),&argMap);err!=nil{
		return nil,err
	}
	if (argMap["storage_fields"] != nil || argMap["leadger_fields"] != nil)&& argMap["schema_auth"] != nil && argMap["auth"] != nil{
		if err := json.Unmarshal([]byte(arg),schema);err!=nil{
			return nil,err
		}
		//检查schema leadger字段
		err := leadgerFieldsCheck(&schema.LeadgerFields)
		if err!=nil{
			return nil,err
		}
		return schema,nil
	}
	for k,v := range argMap{  //假设用户只传递了map类型的信息，我们则默认它全部都是存储字段
		schema.StorageFields[k] = SchemaParameters{
			Value:v,
			PAuth:Auth{},
		}
	}
	return schema,nil
}

func leadgerFieldsCheck(leadgerFields *map[string]SchemaLeadgerParameters)error{
	for k,fields := range *leadgerFields{
		if fields.FCheck == nil {
			fields.FCheck = &Checks{
				FType:"float64",
				Must:false,
				Fieldlen:1,
			}
		}else{
			if fields.FCheck.FType != "float64"{
				return fmt.Errorf("%s fields type must is float64",k)
			}else if !(fields.FCheck.Fieldlen > 0) {
				return fmt.Errorf("%s fields length must is greater than zero",k)
			}
		}
	}
	return nil
}


func fieldCheckout(check *Checks,val interface{})error{
	fType := reflect.TypeOf(val).String()

	switch check.FType {
	case "string":
		if fType != "string"{
			return fmt.Errorf("fields type does not match the defined string type! ")
		}
		if check.Must && (val == ""){
			return fmt.Errorf("fields can not be empty")
		}
		value := val.(string)
		if len([]rune(value))>check.Fieldlen{
			return fmt.Errorf("fields length cannot be greater than %d ",check.Fieldlen)
		}
	case "int","int16","int32","int64","float32","float64":
		if fType != "float64"{
			return fmt.Errorf("fields type does not match the defined %s type! ",check.FType )
		}
		value := val.(float64)
		if check.Must && (value == 0){
			return fmt.Errorf("fields can not be 0")
		}
		if check.Fieldlen !=0  && value < 0{
			return fmt.Errorf("fields cannot be negative")
		}
	case "time.Time":
		if fType != "string"{
			return fmt.Errorf("fields type does not match the defined time.Time type! ")
		}
		dateStr := reflect.ValueOf(val).String()
		if check.Must &&  dateStr == ""{
			return fmt.Errorf("fields cannot be empty")
		}
		if _ ,err := time.Parse("2006-01-02 15:04:05",dateStr);err != nil{
			return fmt.Errorf(err.Error())
		}

		//if check.Must && (dateStr == "0001-01-01 00:00:00 +0000 UTC"|| dateStr == "0001-01-01T00:00:00Z"){
		//	return fmt.Errorf("fields cannot be initial value")
		//}
	default:
		fmt.Errorf("fields type no match! ")
	}
	return nil
}

func Constructor(capacity int) LRUCache{
	lruCache := LRUCache{limit:capacity}
	lruCache.HashMap = make(map[string]*Node, capacity)
	return lruCache
}

func (l *LRUCache) Get(key string) []byte {
	if v,ok:= l.HashMap[key];ok {
		l.refreshNode(v)
		return v.Value
	}else {
		return nil
	}
}

func (l *LRUCache) Put(key string, value []byte) {
	if v,ok := l.HashMap[key];!ok{
		if len(l.HashMap) >= l.limit{
			oldKey := l.removeNode(l.head)
			delete(l.HashMap, oldKey)
		}
		node := Node{Key:key, Value:value}
		l.addNode(&node)
		l.HashMap[key] = &node
	}else {
		v.Value = value
		l.refreshNode(v)
	}
}

func (l *LRUCache) refreshNode(node *Node){
	if node == l.end {
		return
	}
	l.removeNode(node)
	l.addNode(node)
}

func (l *LRUCache) removeNode(node *Node) string{
	if node == l.end  {
		l.end = l.end.pre
		l.end.next = nil
	}else if node == l.head {
		l.head = l.head.next
		l.head.pre = nil
	}else {
		node.pre.next = node.next
		node.next.pre = node.pre
	}
	return node.Key
}

func (l *LRUCache) addNode(node *Node){
	if l.end != nil {
		l.end.next = node
		node.pre = l.end
		node.next = nil
	}
	l.end = node
	if l.head == nil {
		l.head = node
	}
}



func main() {
	err := shim.Start(&GXHCC{})
	if err != nil {
		fmt.Printf("Error starting EncCC chaincode: %s", err)
	}
}
